"""FlaskTester - Pytest fixtures for Flask authenticated internal and external tests.

Pytest: PYTEST_DONT_REWRITE
"""

import os
import io
import re
from typing import Any
import importlib
import logging
import pytest

log = logging.getLogger("flask_tester")


class FlaskTesterError(BaseException):
    """Base exception for FlaskTester package."""
    pass


class AuthError(FlaskTesterError):
    """Authenticator Exception."""
    pass


class Authenticator:
    """Manage authentication for test requests.

    Supported schemes:

    - ``basic``: HTTP Basic Authentication
    - ``param``: password with HTTP or JSON parameters
    - ``bearer``: token in ``Authorization`` *bearer* header
    - ``header``: token in a header
    - ``cookie``: token in a cookie
    - ``tparam``: token in a parameter
    - ``fake``: fake scheme, login directly passed as a parameter

    Constructor parameters:

    - ``allow``: list of allowed schemes.
      default is ``["bearer", "basic", "param"]``
    - ``user``: parameter for user on ``param`` password authentication,
      default is ``USER``
    - ``pwd``: parameter for password on ``param`` password authentication,
      default is ``PASS``
    - ``login``: parameter for user on ``fake`` authentication,
      default is ``LOGIN``
    - ``bearer``: name of bearer scheme for token,
      default is ``Bearer``
    - ``header``: name of header for token,
      default is ``Auth``
    - ``cookie``: name of cookie for token,
      default is ``auth``
    - ``tparam``: name of parameter for token,
      default is ``AUTH``
    """

    _TOKEN_SCHEMES = {"bearer", "header", "cookie", "tparam"}
    _PASS_SCHEMES = {"basic", "param"}

    _AUTH_SCHEMES = {"fake"}
    _AUTH_SCHEMES.update(_TOKEN_SCHEMES)
    _AUTH_SCHEMES.update(_PASS_SCHEMES)

    def __init__(self,
             allow: list[str] = ["bearer", "basic", "param"],
             # parameter names for "basic" and "param"
             user: str = "USER",
             pwd: str = "PASS",
             # parameter name for "fake"
             login: str = "LOGIN",
             # 4 token options
             bearer: str = "Bearer",
             header: str = "Auth",
             cookie: str = "auth",
             tparam: str = "AUTH",
         ):

        self._has_pass, self._has_token = False, False
        for auth in allow:
            assert auth in self._AUTH_SCHEMES
            if auth in self._TOKEN_SCHEMES:
                self._has_token = True
            if auth in self._PASS_SCHEMES:
                self._has_pass = True
        self._allow = allow

        # authentication scheme parameters
        self._user = user
        self._pass = pwd
        self._login = login
        self._bearer = bearer
        self._header = header
        self._cookie = cookie
        self._tparam = tparam

        # password and token credentials
        self._passes: dict[str, str] = {}
        self._tokens: dict[str, str] = {}

    def _set(self, login: str, val: str|None, store: dict[str, str]):
        """Set a key/value in a directory, with None for delete."""
        if val is None:
            if login in store:
                del store[login]
        else:
            assert isinstance(val, str)
            store[login] = val

    def setPass(self, login: str, pw: str|None):
        """Associate a password to a user."""
        if not self._has_pass:
            raise AuthError("cannot set password, no password scheme allowed")
        self._set(login, pw, self._passes)

    def setPasses(self, pws: list[str]):
        """Associate a list of login:password."""
        for lp in pws:
            login, pw = lp.split(":", 1)
            self.setPass(login, pw)

    def setToken(self, login: str, token: str|None):
        """Associate a token to a user."""
        if not self._has_token:
            raise AuthError("cannot set token, no token scheme allowed")
        self._set(login, token, self._tokens)

    def _param(self, kwargs: dict[str, Any], key: str, val: Any):
        """Add request parameter to "json" or "data"."""

        if "json" in kwargs:
            assert isinstance(kwargs["json"], dict)
            kwargs["json"][key] = val
        elif "data" in kwargs:
            assert isinstance(kwargs["data"], dict)
            kwargs["data"][key] = val
        else:
            kwargs["data"] = {key: val}

    def _try_auth(self, auth: str|None, scheme: str) -> bool:
        """Whether to try this authentication scheme."""
        return auth in (None, scheme) and scheme in self._allow

    def setAuth(self, login: str|None, kwargs: dict[str, Any], auth: str|None = None):
        """Set request authentication.

        - login: login target, None means no authentication
        - kwargs: request parameters
        - auth: authentication method, default None is to try allowed schemes, tokens first.
        """

        log.debug(f"setAuth: login={login} auth={auth} allow={self._allow}")

        if login is None:  # not needed
            return

        if auth is not None:
            if auth not in self._AUTH_SCHEMES:
                raise AuthError(f"unexpected auth: {auth}")
            if auth not in self._allow:
                raise AuthError(f"auth is not allowed: {auth}")

        # use token if available and allowed
        if login in self._tokens and auth in (None, "bearer", "header", "cookie", "tparam"):

            token = self._tokens[login]

            if "headers" not in kwargs:
                kwargs["headers"] = {}

            if self._try_auth(auth, "bearer"):
                kwargs["headers"]["Authorization"] = self._bearer + " " + token
            elif self._try_auth(auth, "header"):
                kwargs["headers"][self._header] = token
            elif self._try_auth(auth, "tparam"):
                self._param(kwargs, self._tparam, token)
            elif self._try_auth(auth, "cookie"):
                # FIXME cookie?
                kwargs["headers"]["Cookie"] = self._cookie + "=" + token
            else:
                raise AuthError(f"no token carrier: login={login} auth={auth} allow={self._allow}")

        elif login in self._passes and auth in (None, "basic", "param"):

            if self._try_auth(auth, "basic"):
                kwargs["auth"] = (login, self._passes[login])
            elif self._try_auth(auth, "param"):
                self._param(kwargs, self._user, login)
                self._param(kwargs, self._pass, self._passes[login])
            else:
                raise AuthError(f"no password carrier: login={login} auth={auth} allow={self._allow}")

        elif self._try_auth(auth, "fake"):

            self._param(kwargs, self._login, login)

        else:

            raise AuthError(f"no authentication for login={login} auth={auth} allow={self._allow}")


class RequestFlaskResponse:
    """Wrapper to return a Flask-looking response from a request response.

    This only work for simple responses.

    Available attributes:

    - ``status_code``: integer status code
    - ``data``: body as bytes
    - ``text``: body as a string
    - ``headers``: dict of headers and their values
    - ``cookies``: dict of cookies
    - ``json``: JSON-converted body, or None
    - ``is_json``: whether body was in JSON

    Constructor parameter:

    - ``response``: from request.
    """

    def __init__(self, response):

        self._response = response
        self.status_code = response.status_code
        self.data = response.content
        self.text = response.text
        self.headers = response.headers
        self.cookies = response.cookies
        try:
            self.json = response.json()
            self.is_json = True
        except Exception:
            self.json = None
            self.is_json = False


class Client:
    """Common class for flask authenticated testing.

    Constructor parameters:

    - ``auth`` authenticator
    - ``default_login`` if not ``login`` is set.
    """

    def __init__(self, auth: Authenticator, default_login: str|None = None):
        self._auth = auth
        self._default_login = default_login

    def setToken(self, login: str, token: str|None):
        """Associate a token to a login, *None* to remove."""
        self._auth.setToken(login, token)

    def setPass(self, login: str, password: str|None):
        """Associate a password to a login, *None* to remove."""
        self._auth.setPass(login, password)

    def _request(self, method: str, path: str, **kwargs):
        """Run a request and return response."""
        raise NotImplementedError()

    def request(self, method: str, path: str, status: int|None = None, auth: str|None = None, **kwargs):
        """Run a possibly authenticated HTTP request.

        Mandatory parameters:

        - ``method``: HTTP method ("GET", "POST", "PATCH", "DELETE"…)
        - ``path``: local path under the base URL

        Optional parameters:

        - ``status``: expected HTTP status, *None* to skip status check
        - ``login``: authenticated user, use **explicit** *None* to skip
        - ``auth``: authentication scheme to use
        - ``**kwargs``: more request parameters (headers, data, json…)
        """

        if "login" in kwargs:
            login = kwargs["login"]
            del kwargs["login"]
        else:  # if unset, use default
            login = self._default_login

        self._auth.setAuth(login, kwargs, auth=auth)
        res = self._request(method, path, **kwargs)  # type: ignore

        if status is not None:
            if res.status_code != status:  # show error before aborting
                log.error(f"bad {status} result: {res.status_code} {res.text[:512]}")
            assert res.status_code == status, f"unexpected status {res.status_code}, expecting {status}"

        return res

    def get(self, path, **kwargs):
        """HTTP GET request."""
        return self.request("GET", path, **kwargs)

    def post(self, path, **kwargs):
        """HTTP POST request."""
        return self.request("POST", path, **kwargs)

    def put(self, path, **kwargs):
        """HTTP PUT request."""
        return self.request("PUT", path, **kwargs)

    def patch(self, path, **kwargs):
        """HTTP PATCH request."""
        return self.request("PATCH", path, **kwargs)

    def delete(self, path, **kwargs):
        """HTTP DELETE request."""
        return self.request("DELETE", path, **kwargs)

    def check(self, method: str, path: str, status: int, content: str|None = None, **kwargs):
        """Run a query and check the response.

        Mandatory parameters:

        - ``method``: HTTP method ("GET", "POST", "PATCH", "DELETE"…)
        - ``path``: local path under the base URL
        - ``status``: expected HTTP status

        Optional parameters:

        - ``content``: regular expression in response body
        - ``login``: authenticated user, use explicit *None* to skip
        - ``**kwargs``: more request parameters (headers, data, json…)
        """

        # get HTTP response
        res = self.request(method, path, status=status, **kwargs)

        # check contents
        if content is not None:
            if not re.search(content, res.text, re.DOTALL):
                log.error(f"cannot find {content} in {res.text}")
                assert False, f"expected content {content} not found in {res.text}"

        return res


class RequestClient(Client):
    """Request-based test provider.

    Constructor parameters:

    - ``auth`` authenticator
    - ``base_url`` target server
    - ``default_login`` if not ``login`` is set.
    """

    def __init__(self, auth: Authenticator, base_url: str, default_login=None):
        super().__init__(auth, default_login)
        self._base_url = base_url
        # reuse connections, otherwise it is too slow…
        from requests import Session
        self._requests = Session()

    def _request(self, method: str, path: str, **kwargs):
        """Actual request handling."""

        if "data" in kwargs:
            # "data" to "files" parameter transfer
            data = kwargs["data"]
            files: dict[str, Any] = {}
            for name, whatever in data.items():
                # FIXME what types should be accepted?
                if isinstance(whatever, io.IOBase):
                    files[name] = whatever
                elif isinstance(whatever, tuple):
                    # reorder tuple to match requests expectations:
                    file_handle, file_name, file_type = whatever
                    files[name] = (file_name, file_handle, file_type)
                else:
                    pass
            for name in files:
                del data[name]
            assert "files" not in kwargs
            kwargs["files"] = files
            # sanity
            assert not (files and "json" in kwargs), "cannot mix file upload and json?"

        res = self._requests.request(method, self._base_url + path, **kwargs)

        return RequestFlaskResponse(res)


class FlaskClient(Client):
    """Flask-based test provider.

    Constructor parameters:

    - ``auth`` authenticator
    - ``client`` Flask ``test_client``
    - ``default_login`` if not ``login`` is set.
    """

    def __init__(self, auth: Authenticator, client, default_login=None):
        super().__init__(auth, default_login)
        self._client = client

    def _request(self, method: str, path: str, **kwargs):
        """Actual request handling."""
        return self._client.open(method=method, path=path, **kwargs)


@pytest.fixture
def ft_authenticator():
    """Pytest Fixture: ft_authenticator.

    Environment variables:

    - ``FLASK_TESTER_LOG_LEVEL``: package log level in
      ``DEBUG INFO WARNING ERROR CRITICAL NOSET``.
    - ``FLASK_TESTER_ALLOW``: allowed space-separated authentication schemes, in
      ``basic param bearer header cookie tparam fake``.
      Default is ``bearer basic param``.
    - ``FLASK_TESTER_USER``: user login parameter for ``param`` authentication.
      Default is ``USER``.
    - ``FLASK_TESTER_PASS``: user password parameter for ``param`` authentication.
      Default is ``PASS``.
    - ``FLASK_TESTER_LOGIN``: user login parameter for ``fake`` authentication.
      Default is ``LOGIN``.
    - ``FLASK_TESTER_BEARER``: bearer name for *token* authentication.
      Default is ``Bearer``.
    - ``FLASK_TESTER_HEADER``: header name for *token* authentication.
      Default is ``Auth``.
    - ``FLASK_TESTER_COOKIE``: cookie name for *token* authentication.
      Default is ``auth``.
    - ``FLASK_TESTER_TPARAM``: parameter for *token* authentication.
      Default is ``AUTH``.
    - ``FLASK_TESTER_AUTH``: initial comma-separated list of *login:password*.
    """

    level = os.environ.get("FLASK_TESTER_LOG_LEVEL", "NOTSET")
    log.setLevel(logging.DEBUG if level == "DEBUG" else
                 logging.INFO if level == "INFO" else
                 logging.WARNING if level == "WARNING" else
                 logging.ERROR if level == "ERROR" else
                 logging.CRITICAL if level == "CRITICAL" else
                 logging.NOTSET)

    allow = os.environ.get("FLASK_TESTER_ALLOW", "bearer basic param").split(" ")

    # per-scheme parameters, must be consistent with FSA configuration
    user = os.environ.get("FLASK_TESTER_USER", "USER")
    pwd = os.environ.get("FLASK_TESTER_PASS", "PASS")
    login = os.environ.get("FLASK_TESTER_LOGIN", "LOGIN")
    bearer = os.environ.get("FLASK_TESTER_BEARER", "Bearer")
    header = os.environ.get("FLASK_TESTER_HEADER", "Auth")
    cookie = os.environ.get("FLASK_TESTER_COOKIE", "auth")
    tparam = os.environ.get("FLASK_TESTER_TPARAM", "AUTH")

    # create authenticator, possibly with initial credentials
    auth = Authenticator(allow, user=user, pwd=pwd, login=login, bearer=bearer, header=header, cookie=cookie, tparam=tparam)

    if "FLASK_TESTER_AUTH" in os.environ:
        auth.setPasses(os.environ["FLASK_TESTER_AUTH"].split(","))

    yield auth


@pytest.fixture
def ft_client(ft_authenticator):
    """Pytest Fixture: ft_client.

    Target environment variable, one **must** be defined:

    - ``FLASK_TESTER_URL``: application HTTP base URL.
    - ``FLASK_TESTER_APP``: Flask application, eg ``app:create_app``.

    Other environment variable:

    - ``FLASK_TESTER_DEFAULT``: Default client login, default is *None* for no
      default.
    """

    default_login = os.environ.get("FLASK_TESTER_DEFAULT", None)
    client: Client

    if "FLASK_TESTER_URL" in os.environ:  # pragma: no cover

        app_url = os.environ["FLASK_TESTER_URL"]
        client = RequestClient(ft_authenticator, app_url, default_login)

    elif "FLASK_TESTER_APP" in os.environ:

        # load app package
        pkg_name, app = os.environ["FLASK_TESTER_APP"], None
        app_names = ["app", "application", "create_app", "make_app"]
        if ":" in pkg_name:  # override defaults
            pkg_name, app_name = pkg_name.split(":", 1)
            app_names = [app_name]
        pkg = importlib.import_module(pkg_name)
        # find app in package
        for name in app_names:
            if hasattr(pkg, name):
                app = getattr(pkg, name)
                if callable(app) and not hasattr(app, "test_client"):
                    app = app()
                break
        if not app:  # pragma: no cover
            raise FlaskTesterError(f"cannot find Flask app in {pkg_name}")
        client = FlaskClient(ft_authenticator, app.test_client(), default_login)

    else:  # pragma: no cover

        raise FlaskTesterError("no Flask application to test")

    yield client
